/*****************************************************************************************************************************
--
--    File Name    : embedded_sync.v

--    Description  : This module implements downsampler for YCbCr 4:4:4 to 4:2:2--                                  

-- Targeted device : Microsemi-SoC                     
-- Author          : India Solutions Team

-- SVN Revision Information:
-- SVN $Revision: TBD
-- SVN $Date: TBD
--
--
--
-- COPYRIGHT 2016 BY MICROSEMI 
-- THE INFORMATION CONTAINED IN THIS DOCUMENT IS SUBJECT TO LICENSING RESTRICTIONS 
-- FROM MICROSEMI CORP.  IF YOU ARE NOT IN POSSESSION OF WRITTEN AUTHORIZATION FROM 
-- MICROSEMI FOR USE OF THIS FILE, THEN THE FILE SHOULD BE IMMEDIATELY DESTROYED AND 
-- NO BACK-UP OF THE FILE SHOULD BE MADE. 
-- 

--******************************************************************************************************************************/
`timescale 1ns/1ps

module embedded_sync #(
/******************************************************************************************************************************/
/***********************************************Parameters*********************************************************************/
/******************************************************************************************************************************/
                        parameter g_DATAWIDTH  = 8
                     )
                     (
/******************************************************************************************************************************/
/***********************************************Module IO**********************************************************************/
/******************************************************************************************************************************/  
                        //System Clock
                        input wire                       CLOCK_I,
                        //System Reset                        
                        input wire                       RESET_N_I,
                        //Luminance data input
                        input wire [(g_DATAWIDTH-1) : 0] Y_in_i,
                        //Chroma Data Input
                        input wire [(g_DATAWIDTH-1) : 0] C_in_i,
                        //Data Input enable
                        input wire                       data_enable_i,
                        //Horizontal Sync
                        input wire                       horz_sync_i,
                        //Vertical Sync
                        input wire                       vert_sync_i,
                        //Horizontal data active signal
                        input wire                       hactive_i,
                        //Vertical data active signal
                        input wire                       vactive_i,
                        //Luminance data output
                        output reg [(g_DATAWIDTH-1) : 0] Y_out_o,
                        //Chroma Data Output
                        output reg [(g_DATAWIDTH-1) : 0] C_out_o,
                        //Data Input enable
                        output reg                       data_enable_o,
                        //Horizontal Sync
                        output reg                       horz_sync_o,
                        //Vertical Sync
                        output reg                       vert_sync_o                        
                     );
/******************************************************************************************************************************/
/*****************************Local Parameters Declarations********************************************************************/
/******************************************************************************************************************************/
  localparam PIPELINE_STAGE = 6;
  localparam IDLE    = 3'd0;
  localparam FF_ST   = 3'd1;
  localparam ST_00_1 = 3'd2;
  localparam ST_00_2 = 3'd3;
  localparam ST_CW   = 3'd4;
  localparam ST_ACT  = 3'd5;

  reg[2 : 0] state,
             next;

  reg [(g_DATAWIDTH-1) : 0] Y_reg[(PIPELINE_STAGE-1) : 0];
  reg [(g_DATAWIDTH-1) : 0] C_reg[(PIPELINE_STAGE-1) : 0];
  reg h_reg[(PIPELINE_STAGE-1) : 0];
  reg v_reg[(PIPELINE_STAGE-1) : 0];
  reg sync_p[(PIPELINE_STAGE-1) : 1];
  reg hsync_d1;
  reg hsync_d2;
  reg vsync_d1;
  reg vsync_d2;
  reg dvi_de_d1;
  reg dvi_de_d2;

/******************************************************************************************************************************/
/********************************Process to register all the inputs************************************************************/
/******************************************************************************************************************************/
  always
    @(posedge CLOCK_I or
      negedge RESET_N_I)
    begin: LATCH_INPUTS
      if (~RESET_N_I)
        begin
          Y_reg[0]  <= { g_DATAWIDTH { 1'b0 } };
          C_reg[0]  <= { g_DATAWIDTH { 1'b0 } };
          h_reg[0]  <= 1'b0;
          v_reg[0]  <= 1'b0;
          hsync_d1  <= 1'b0;
          vsync_d1  <= 1'b0;
          dvi_de_d1 <= 1'b0;
        end
      else
        begin
          Y_reg[0] <= Y_in_i;
          C_reg[0] <= C_in_i;
          h_reg[0] <= hactive_i;
          v_reg[0] <= vactive_i;
          hsync_d1 <= horz_sync_i;
          vsync_d1 <= vert_sync_i;
          dvi_de_d1 <= data_enable_i;
        end
    end
/******************************************************************************************************************************/
/********************************Process to insert required pipeline stages****************************************************/
/******************************************************************************************************************************/
  always
    @(posedge CLOCK_I or
      negedge RESET_N_I)
    begin: PIPELINE_STAGES
      integer i,
              j;
      if (~RESET_N_I)
        begin
          for (i = 1; i < PIPELINE_STAGE; i = i + 1)
            begin
              Y_reg[i] <= 8'h0;
              C_reg[i] <= 8'h0;
              h_reg[i] <= 1'b0;
              v_reg[i] <= 1'b0;
              sync_p[i] <= 1'b0;
              hsync_d2 <= 1'b0;
              vsync_d2 <= 1'b0;
              dvi_de_d2 <= 1'b0;
            end
        end
      else
        begin
          for (i = 1; i < PIPELINE_STAGE; i = i + 1)
            begin
              Y_reg[i] <= Y_reg[i - 1];
              C_reg[i] <= C_reg[i - 1];
              h_reg[i] <= h_reg[i - 1];
              v_reg[i] <= v_reg[i - 1];
              // sync_p[i]<= (h_reg[i]^h_reg[i-1] | v_reg[i]^v_reg[i-1]);
              sync_p[i] <= (h_reg[i] ^ h_reg[i - 1]);
              hsync_d2 <= hsync_d1;
              vsync_d2 <= vsync_d1;
              dvi_de_d2 <= dvi_de_d1;
            end
        end
    end
/******************************************************************************************************************************/
/***********************************************MAIN FSM***********************************************************************/
/******************************************************************************************************************************/
  always
    @*
    case (state)
      IDLE:
        if (sync_p[1])
          next = FF_ST;
        else
          next = IDLE;
      FF_ST:
        next = ST_00_1;
      ST_00_1:
        next = ST_00_2;
      ST_00_2:
        next = ST_CW;
      ST_CW:
        if (h_reg[0])
          next = ST_ACT;
        else
          next = IDLE;
      ST_ACT:
        if (sync_p[5])
          next = FF_ST;
        else
          next = ST_ACT;
      default:
        next = IDLE;
    endcase

  always
    @(posedge CLOCK_I or
      negedge RESET_N_I)
    if (~RESET_N_I)
      state <= IDLE;
    else
      state <= next;
/******************************************************************************************************************************/
/***********************************************Process ro insert embedded sync codes******************************************/
/******************************************************************************************************************************/
  always
    @(posedge CLOCK_I or
      negedge RESET_N_I)
    begin: EMBED_SYNC
      if (~RESET_N_I)
        begin
          Y_out_o <= 8'h10;
          C_out_o <= 8'h80;
          horz_sync_o <= 1'b0;
          vert_sync_o <= 1'b0;
          data_enable_o <= 1'b0;
        end
      else
        begin
          horz_sync_o <= hsync_d2;
          vert_sync_o <= vsync_d2;
          data_enable_o <= dvi_de_d2;
          case (next)
            IDLE:
              begin
                Y_out_o <= 8'h10;
                C_out_o <= 8'h80;
              end
            FF_ST:
              begin
                Y_out_o <= 8'hFF;
                C_out_o <= 8'hFF;
              end
            ST_00_1:
              begin
                Y_out_o <= 8'h00;
                C_out_o <= 8'h00;
              end
            ST_00_2:
              begin
                Y_out_o <= 8'h00;
                C_out_o <= 8'h00;
              end
            ST_CW:
              begin
                case ({ ~v_reg[3], ~h_reg[3] })
                  2'b00:
                    begin
                      { Y_out_o, C_out_o } <= { 8'h80, 8'h80 }; // SAV Active video
                    end
                  2'b01:
                    begin
                      { Y_out_o, C_out_o } <= { 8'h9D, 8'h9D }; // EAV Active video
                    end
                  2'b10:
                    begin
                      { Y_out_o, C_out_o } <= { 8'hAB, 8'hAB }; // SAV blanking 
                    end
                  2'b11:
                    begin
                      { Y_out_o, C_out_o } <= { 8'hB6, 8'hB6 }; // EAV blanking
                    end
                endcase
              end
            ST_ACT:
              begin
                Y_out_o <= Y_reg[5];
                C_out_o <= C_reg[5];
              end
          endcase
        end
    end
endmodule
